# (c) Copyright 2009. CodeWeavers, Inc.

import urllib2
import traceback

import cxlog

def _get_rfc822_value(string):
    if string.startswith('"'):
        # quoted-string
        return string.split('"', 2)[1]
    else:
        # token
        def is_token_breaker(c):
            return c in ' ()<>@,;:\\"/[]?=' or not (0x1f < ord(c) < 0x7f)
        chars = []
        for c in string:
            if is_token_breaker(c):
                break
            chars.append(c)
        return ''.join(chars)

def _replace_invalid_chars(string):
    chars = []
    for c in string:
        if c in '\\:*?"<>|\x7f' or ord(c) < 0x20:
            chars.append('_')
        else:
            chars.append(c)
    return ''.join(chars)

def _basename(string):
    result = string.rsplit('/', 1)[-1]
    return _replace_invalid_chars(result)

class StopDownload(Exception):
    "Stops the download if raised from a notify_progress handler."
    pass

class UrlGetter(object):
    bytes_downloaded = 0
    bytes_total = -1
    finished = False
    url = ''
    basename = ''
    outfile = None
    infile = None
    notify_progress = None
    notify_finished = None
    notify_failed = None

    blocksize = 4096

    def _do_nothing(self, *args):
        pass

    def __init__(self, url, outfile, notify_progress=_do_nothing, notify_finished=_do_nothing, notify_failed=_do_nothing):
        self.url = url
        self.basename = _basename(url)
        if hasattr(outfile, 'write'):
            self.outfile = outfile
        else:
            self.outfile = open(outfile, 'wb')
        self.notify_progress = notify_progress
        self.notify_finished = notify_finished
        self.notify_failed = notify_failed

    def fetch(self):
        if self.infile:
            raise Exception("fetch() has already been called")
        cxlog.log("fetching url %s" % cxlog.to_str(self.url))
        try:
            try:
                self.infile = urllib2.urlopen(self.url)
                headers = self.infile.info()

                # calculate the basename
                basename = ''
                if 'filename=' in headers.get('content-disposition', ()):
                    # FIXME: This is not entirely correct because filename= could appear in some other value
                    _unused, filenamevalue = headers['content-disposition'].split('filename=', 1)
                    basename = _basename(_get_rfc822_value(filenamevalue))
                if not basename:
                    # recheck the url in case we were redirected
                    basename = _basename(self.infile.geturl())
                self.basename = basename

                if 'content-length' in headers:
                    self.bytes_total = int(headers['content-length'])
                try:
                    self.notify_progress(self)
                except StopDownload:
                    return

                while True:
                    data = self.infile.read(self.blocksize)
                    if data == '':
                        break
                    else:
                        self.outfile.write(data)
                        # Not += to avoid ambiguity with the static variable
                        self.bytes_downloaded = self.bytes_downloaded + len(data)
                        try:
                            self.notify_progress(self)
                        except StopDownload:
                            return
                if self.bytes_total != -1 and self.bytes_downloaded != self.bytes_total:
                    raise Exception("expected to download %i bytes, got %i" % (self.bytes_total, self.bytes_downloaded))
                self.finished = True
                cxlog.log("fetched url %s (%i bytes)\n" % (cxlog.to_str(self.url), self.bytes_downloaded))
                self.notify_finished(self)
            except Exception, exception:
                cxlog.err("failed to fetch url %s:\n%s" % (cxlog.to_str(self.url), traceback.format_exc()))
                self.notify_failed(self, exception)
        finally:
            self.outfile.close()

def main():
    import sys
    def notify_progress(urlgetter):
        print '%s/%s bytes' % (urlgetter.bytes_downloaded, urlgetter.bytes_total)
    def notify_finished(_urlgetter):
        print 'Success.'
    def notify_failed(_urlgetter, _exception):
        print 'Failure.'
    getter = UrlGetter(sys.argv[1], sys.argv[2], notify_progress, notify_finished, notify_failed)
    getter.fetch()
    print "suggested basename:", getter.basename

if __name__ == '__main__':
    main()

