# (c) Copyright 2009-2010. CodeWeavers, Inc.

# for localization
from cxutils import cxgettext as _

import os
import traceback

import gtk
import gtk.glade
import gobject
gtk.gdk.threads_init()

import cxutils
import bottlequery
import cxlog
import pyop

import c4profiles
import c4profilesmanager
import installtask
import cxaiebase
import cxaiecore
import cxaiengine
import cxaiemedia

import cxguitools
import bottlecollection
import mountpoints


def package_select_filter(model, filteriter, wizard):
    isGroup = model.get_value(filteriter,3)
    if isGroup or model.get_value(filteriter, 1) == 'com.codeweavers.unknown':
        return True
    packageName = model.get_value(filteriter, 0)
    searchString = wizard.packageSearchString
    if searchString and packageName:
        count = packageName.lower().count(searchString.lower())
        if count > 0:
            return True 
        else:
            return False
    else:
        return True
    

class InstallerAssistantController(installtask.InstallTaskDelegate):

    def __init__(self, **kwargs):
        installtask.InstallTaskDelegate.__init__(self)
        self.options = kwargs

        #-----------------------------------------------------------------------
        #  Setup the GUI
        #-----------------------------------------------------------------------

        try:
            self.gladepath = os.environ["CX_GLADEPATH"]
        except KeyError:
            self.gladepath = os.path.join(cxutils.CX_ROOT, "lib", "python", "glade")

        self.gladefile = os.path.join(self.gladepath, "installerassistant.glade")

        gtk.glade.set_custom_handler(cxguitools.CustomWidgets.custom_widget_handler)
        if not gtk.glade.textdomain() == "crossover":
            locale_path = os.path.join(cxutils.CX_ROOT, "share", "locale")
            gtk.glade.bindtextdomain("crossover", locale_path)
            gtk.glade.textdomain("crossover")
        self.xml = gtk.glade.XML(self.gladefile)
        self.xml.signal_autoconnect(self)
        self.animateFunctions = []

        self.installTask = installtask.InstallTask(delegate=self)
        self.installEngine = None
        self.install_in_progress = False
        self.install_ui = False # True if we're doing something in an install that requires user interaction
        self.failing_tasks = []

        # Start parsing any extra package information now
        if self.options.get('c4pfile'):
            self.parse_c4pfile(self.options['c4pfile'])

        self.autorun_file = None

        # Tweaks not allowed by glade.
        install_notes = self.xml.get_widget('InstallationNotes')
        install_notes.set_size_request(1, -1)
        warning_image = install_notes.render_icon(gtk.STOCK_DIALOG_WARNING, gtk.ICON_SIZE_MENU)
        if warning_image:
            install_notes.images[u'warning'] = warning_image
        error_image = install_notes.render_icon(gtk.STOCK_DIALOG_ERROR, gtk.ICON_SIZE_MENU)
        if error_image:
            install_notes.images[u'error'] = error_image

        self.xml.get_widget('PackageDescription').set_size_request(1,-1)

        self.xml.get_widget('DownloadInstallationNotes').set_size_request(1,-1)

        self.xml.get_widget('InstallResults').set_size_request(1,-1)


        # The fields are: label, icon, mountpoint, sort key
        self.volumesListModel = gtk.ListStore(str, gtk.gdk.Pixbuf, str, str)
        self.volumesListModel.set_sort_column_id(3, gtk.SORT_ASCENDING)

        volumesWidget = self.xml.get_widget("MediaTreeView")
        volumesWidget.set_text_column(0)
        volumesWidget.set_pixbuf_column(1)
        volumesWidget.set_model(self.volumesListModel)

        self.mountpoints = mountpoints.MountPointsNotifier(self.volumeAdded, self.volumeDeleted)

        # Add the special 'Browse...' options
        newrow = self.volumesListModel.append()
        self.volumesListModel.set_value(newrow, 0, self.chooseInstallerFileString)
        pixbuf = volumesWidget.render_icon(gtk.STOCK_EXECUTE, gtk.ICON_SIZE_DIALOG)
        self.volumesListModel.set_value(newrow, 1, pixbuf)
        self.volumesListModel.set_value(newrow, 3, '1')

        newrow = self.volumesListModel.append()
        self.volumesListModel.set_value(newrow, 0, self.chooseInstallerFolderString)
        pixbuf = volumesWidget.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_DIALOG)
        self.volumesListModel.set_value(newrow, 1, pixbuf)
        self.volumesListModel.set_value(newrow, 3, '2')

        bottlesWidget = self.xml.get_widget("BottleSelectionView")
        column = gtk.TreeViewColumn(_("Name"), gtk.CellRendererText(), markup = 1)
        column.set_resizable(True)
        bottlesWidget.append_column(column)
        column = gtk.TreeViewColumn(_("Type"), gtk.CellRendererText(), text = 2)
        column.set_resizable(False)
        bottlesWidget.append_column(column)
        self.bottleListModel = gtk.TreeStore(str, str, str, str) #name, display markup, template, sort key
        self.bottleListModel.set_sort_column_id(3, gtk.SORT_ASCENDING)
        bottlesWidget.set_model(self.bottleListModel)

        downloadCombo = self.xml.get_widget("DownloadCombo")
        self.urlListStore = gtk.ListStore(str, str, str)
        downloadCombo.set_model(self.urlListStore)

        # The fields are: name, appid, sortkey
        self.packageTreeStore = gtk.TreeStore(str, str, str, bool)
        self.packageTreeModel = self.packageTreeStore.filter_new()
        packageTreeView = self.xml.get_widget("PackageTreeView")
        packageTreeView.set_model(self.packageTreeModel)
        # The column header is not shown so there is no point localizing it
        column = gtk.TreeViewColumn("Installable Packages", gtk.CellRendererText(), markup = 0)
        column.set_resizable(False)
        packageTreeView.append_column(column)
        self.packageTreeStore.set_sort_column_id(2, gtk.SORT_ASCENDING)
        self.packageTreeModel.set_visible_func(package_select_filter,self)
        self.packageSearchString = ""

        for bottle in bottlecollection.sharedCollection().bottles():
            self.installTask.add_bottle(bottle)

        bottlecollection.waitForSharedCollectionToBeReady(self.show_dialog, ("basic",))

        gobject.timeout_add(100, self.animate)

        self.bottle_progress_source = gobject.timeout_add(100, self.update_bottle_analysis_progress)

    def parseResponse(self, response, userdata):
        if response == 0:
            self.autorun_file = c4profilesmanager.C4ProfilesSet.add_c4p_file(userdata,False)
            self.installTask.update_profiles()
            if self.xml.get_widget('InstallerAssistant').get_property('visible'):
                self.updatePackageListing()
                self.installTask.use_autorun_file(self.autorun_file)
                self.installTask.AutoFillSettings()

    def is_c4pfile_trusted(self, _c4pfile):
        # at present all external c4pfiles are treated as untrusted
        return False

    def parse_c4pfile(self, c4pfile):
        trusted = self.is_c4pfile_trusted(c4pfile)
        if not trusted:
            title = _("Load untrusted C4P file?")
            msg = _("The C4P file '%(filename)s' is not from a trusted source!\n\nThis means that CodeWeavers has no knowledge or control over what actions profiles from this C4P file may attempt. Be VERY cautious loading untrusted C4Ps as they can cause great damage to your computer.\n\nFinally, C4P loading is still in alpha and may be unreliable. You will get no support from CodeWeavers with issues involved in loading untrusted C4P files.\n\n Despite all of this, do you want to proceed?") %{'filename':c4pfile}
            cxguitools.CXMessageDlg(msg, title=title,
                                        button_array=[[_("Don't Load"), 1],
                                                      [_("Load"), 0]],
                                        user_data = c4pfile,
                                        response_function=self.parseResponse)
        else:
            self.autorun_file = c4profilesmanager.C4ProfilesSet.add_c4p_file(c4pfile, False)

    def cancel_clicked(self, _caller):
        if self.install_in_progress:
            self.cancel_install()
        else:
            self.destroy()

    def destroy(self):
        if self.bottle_progress_source:
            gobject.source_remove(self.bottle_progress_source)
            self.bottle_progress_source = None
        self.xml.get_widget("InstallerAssistant").destroy()
        self.quit_requested(None)

    def quit_requested(self, _caller):
        cxguitools.toplevel_quit()

    def installed_applications_ready(self):
        if not self.installEngine:
            for bottle in bottlecollection.sharedCollection().bottles():
                self.installTask.installed_applications_ready(bottle)

    def show_dialog(self):
        dialog = self.xml.get_widget("InstallerAssistant")
        self.updatePackageListing()
        self.profileChanged()
        self.updateBottleListing()
        self.sourceChanged()
        self.bottleCreateChanged()
        self.bottleNewnameChanged()
        self.bottleTemplateChanged()

        bottlecollection.waitForSharedCollectionToBeReady(self.installed_applications_ready, ('applications',))

        if self.options.get('bottle') and self.options['bottle'] in bottlecollection.sharedCollection().bottleList():
            self.installTask.bottlename = self.options['bottle']

        if self.options.get('installersource'):
            self.installTask.SetInstallerSource(self.options['installersource'])

        if self.options.get('profileid') and self.options['profileid'] in self.installTask.profiles:
            self.installTask.profile = self.installTask.profiles[self.options['profileid']]

        if self.autorun_file:
            self.installTask.use_autorun_file(self.autorun_file)

        # Try to guess the profile from inserted CD's.
        self.installTask.AutoFillSettings()

        self.set_expanded_section('PackageSelect')
        dialog.show()

    def set_expanded_section(self, name):
        for section_name in ['PackageSelect', 'MediaSelect', 'BottleSelect', 'Installer', 'Finished']:
            if section_name == name:
                self.xml.get_widget('%sArrow' % section_name).set(gtk.ARROW_DOWN, gtk.SHADOW_NONE)
                self.xml.get_widget('%sContent' % section_name).show()
            else:
                self.xml.get_widget('%sArrow' % section_name).set(gtk.ARROW_RIGHT, gtk.SHADOW_NONE)
                self.xml.get_widget('%sContent' % section_name).hide()
        self.on_section_changed[name](self)

    def advance_section(self, can_install=False):
        # Expand the first section that needs work.
        # If can_install is True and no sections need work, start the install.
        if self.installEngine:
            # Install has started already.
            self.destroy()
        elif self.wantPackageSelection():
            self.set_expanded_section('PackageSelect')
        elif self.wantMediaSelection():
            self.set_expanded_section('MediaSelect')
        elif self.wantBottleSelection():
            self.set_expanded_section('BottleSelect')
        elif can_install and self.installTask.can_install():
            self.displayInstallationProgressPanel()
            self.doInstall()
        else:
            self.set_expanded_section(None)

    def next_clicked(self, _widget):
        self.advance_section(True)

    def update_common_widgets(self):
        # update next button
        nextButton = self.xml.get_widget("InstallerNext")
        if (self.xml.get_widget('PackageSelectContent').get_property('visible') and self.wantPackageSelection() or
            self.xml.get_widget('MediaSelectContent').get_property('visible') and self.wantMediaSelection() or
            self.xml.get_widget('BottleSelectContent').get_property('visible') and self.wantBottleSelection()):
            nextButton.set_label(_("Proceed"))
            nextButton.set_sensitive(False)
        elif self.wantPackageSelection() or self.wantMediaSelection() or self.wantBottleSelection():
            nextButton.set_label(_("Proceed"))
            nextButton.set_sensitive(True)
        elif not self.installTask.can_install():
            # Can't install but no sections need work. Install notes should say why.
            nextButton.set_label(_("Proceed"))
            nextButton.set_sensitive(False)
        else:
            nextButton.set_label(_("Install"))
            nextButton.set_sensitive(True)

        self.update_install_notes(True)

    def update_install_notes(self, show_warnings):
        installationNotesTextView = self.xml.get_widget("InstallationNotes")

        installationNotesText = []

        warning_icon_html = '<img src="warning" alt="WARNING" />'
        error_icon_html = '<img src="error" alt="ERROR" />'

        if self.installTask.profile is None:
            installationNotesText.append(_("Please select a software package to install."))
        elif self.installTask.installationNotes:
            installationNotesText.append(self.installTask.installationNotes)

        if show_warnings:
            if self.installTask.target_bottle and \
                    self.installTask.target_bottle.analyzed():
                for warning in self.installTask.target_bottle.warnings:
                    installationNotesText.append(warning_icon_html + " " + warning)

            # cxdiag messages
            for level, title, description in self.installTask.get_cxdiag_messages().values():
                if level == 'Require':
                    installationNotesText.append('%s %s: %s' % (warning_icon_html, title, description))

            # check that the new bottle name doesn't match an existing bottle
            if self.installTask.GetCreateNewBottle() and self.installTask.bottlename:
                for bottlename in bottlecollection.sharedCollection().bottleList():
                    if self.installTask.bottlename.lower() == bottlename.lower():
                        installationNotesText.append(_("%(error_icon)s There is already a bottle named '%(bottlename)s'. Select the existing bottle from the list or choose a different bottle name.") % {
                            'error_icon': error_icon_html,
                            'bottlename': self.installTask.bottlename})
                        break

            if not self.wantPackageSelection() and not self.wantMediaSelection() and not self.wantBottleSelection():
                # We have a greyed out Install button
                if not self.installTask.target_bottle.analyzed():
                    installationNotesText.append(_("The install cannot continue until the '%s' bottle is scanned for dependencies. Please wait.") % self.installTask.bottlename)

            if self.installTask.target_bottle and self.installTask.target_bottle.analyzed() and self.installTask.target_bottle.missing_media:
                installationNotesText.append(error_icon_html + u' ' + _("The install cannot continue because one or more dependencies are not installed, and an installer could not be found automatically."))

        if installationNotesText:
            installationNotesTextView.get_buffer().set_text("")
            htmlNotes = '<body>%s</body>' % (''.join('<p>%s</p>' % x for x in installationNotesText))
            installationNotesTextView.display_html(htmlNotes.encode("utf8"))
            installationNotesTextView.show()
        else:
            installationNotesTextView.hide()


    #-----------------------------------------------------------------------
    #  Package Selection
    #-----------------------------------------------------------------------
    def wantPackageSelection(self):
        return self.installTask.profile is None

    def package_toggle(self, _widget):
        if self.xml.get_widget('PackageSelectContent').get_property('visible'):
            self.advance_section()
        else:
            self.set_expanded_section('PackageSelect')

    communitySupportedString = "<b>"+_("Community Supported Applications")+"</b>"
    unsupportedString = "<b>"+_("Unsupported Applications")+"</b>"
    supportedString = "<b>"+_("Supported Applications")+"</b>"
    componentString = "<b>"+_("Runtime Support Components")+"</b>"
    upgradeString = "<b>"+_("Service Packs and Upgrades")+"</b>"
    def updatePackageListing(self):
        packageTreeView = self.xml.get_widget("PackageTreeView")
        self.packageTreeStore.clear()

        supportedParent = self.packageTreeStore.append(None, [self.supportedString, '.supported', "zzzz1 " + self.supportedString, True])
        unsupportedParent = self.packageTreeStore.append(None, [self.unsupportedString, '.unsupported', "zzzzz3 " + self.unsupportedString, True])
        communityParent = self.packageTreeStore.append(unsupportedParent, [self.communitySupportedString, '.communitysupported', self.communitySupportedString, True])
        upgradeParent = self.packageTreeStore.append(None, [self.upgradeString, '.upgrade', "zzzzz4 " + self.upgradeString, True])
        componentParent = self.packageTreeStore.append(None, [self.componentString, '.component', "zzzzz5 " + self.componentString, True])
        for appid, profile in self.installTask.profiles.iteritems():
            if not profile.is_for_current_product:
                continue
            # FIXME: If app_profile.extra_fors matches one of the installed
            # applications, then put it at the top of the list.
            if profile.app_profile and profile.app_profile.category == "Component/Service Pack":
                self.packageTreeStore.append(upgradeParent, [profile.name, appid, profile.name, False])
            else:
                if appid == c4profilesmanager.UNKNOWN:
                    self.packageTreeStore.append(unsupportedParent, [profile.name, appid, profile.name, False])
                elif profile.app_profile.is_supported:
                    if profile.is_application:
                        self.packageTreeStore.prepend(supportedParent, [profile.name, appid, profile.name, False])
                    if profile.is_component:
                        self.packageTreeStore.append(componentParent, [profile.name, appid, profile.name, False])
                    if not profile.is_component and not profile.is_application:
                        self.packageTreeStore.append(supportedParent, [profile.name, appid, profile.name, False])
                else:
                    self.packageTreeStore.append(communityParent, [profile.name, appid, profile.name, False])
        packageTreeView.expand_row(0,False)
        packageTreeView.expand_row(1,False)
        packageTreeView.set_cursor(0)

        installerSourceWidgets = self.xml.get_widget("PackageInstallerSourceWidgets")
        if self.installTask and self.installTask.GetInstallerSource():
            installerSourceWidgets.show()
            installerSourceLabel = self.xml.get_widget("PackageInstallerSourceLabel")
            installerSourceLabel.set_text(self.installTask.GetInstallerSource())
        else:
            installerSourceWidgets.hide()

    def package_changed_gui(self, _widget):
        packageTreeView = self.xml.get_widget("PackageTreeView")
        cursor = packageTreeView.get_cursor()
        if cursor and cursor[0]:
            is_category = self.packageTreeModel.get_value(self.packageTreeModel.get_iter(cursor[0]), 3)
            if not is_category:
                selectedID = self.packageTreeModel.get_value(self.packageTreeModel.get_iter(cursor[0]), 1)
                self.installTask.profile = self.installTask.profiles[selectedID]
                if cxlog.is_on('c4profiles'):
                    cxlog.log_('c4profiles', "selected %s" % cxlog.to_str(self.installTask.profile.appid))
                    self.installTask.profile.dump(cxlog.get_file(), "| ")
            else:
                self.installTask.profile = None

    def select_package_by_appid(self, model, path, treeiter, appid):
        if model.get_value(treeiter, 1) == appid:
            self.xml.get_widget("PackageTreeView").set_cursor(path)
            return True # stop iteration
        else:
            return False # continue iteration

    medal_names = {
        'gold': _("Gold Medal (Officially Supported)"),
        'silver': _("Silver Medal (Officially Supported)"),
        'bronze': _("Bronze Medal (Officially Supported)"),
        'ungold': _("Gold Medal (Unsupported)"),
        'unsilver': _("Silver Medal (Unsupported)"),
        'unbronze': _("Bronze Medal (Unsupported)"),
        'knownnottowork': _("Known not to work"),
        'untested': _("Untested"),
    }

    medal_descriptions = {
        'gold': _("Installs and runs as you would expect in Microsoft Windows. We expect that you can use this application on an everyday basis with good results and only minor bugs."),
        'silver': _("Installs and runs well enough to be usable. However, this application has bugs that prevent it from running flawlessly."),
        'bronze': _("Installs, runs, and can accomplish some portion of its fundamental mission. However, this application has enough bugs that we recommend you use it with caution. Save early/save often, and don't be surprised if there are some bumps along the way."),
        'ungold': _("Installs and runs as you would expect in Microsoft Windows. Our users expect that you can use this application on an everyday basis with good results and only minor bugs."),
        'unsilver': _("Installs and runs well enough to be usable. However, our users find that this application has bugs that prevent it from running flawlessly."),
        'unbronze': _("Installs, runs, and can accomplish some portion of its fundamental mission. However, this application has enough bugs that we recommend you use it with caution. Save early/save often, and don't be surprised if there are some bumps along the way."),
        'knownnottowork': _("This application either does not install or does not run."),
        'untested': _("We do not know whether or not this application will install and run. If you try it, consider letting us know how it runs, using the compatibility database."),
    }

    def profileChanged(self):
        supportedLabel = self.xml.get_widget("PackageSupportedLabel")
        medalImage = self.xml.get_widget("MedalImage")
        descriptionTextView = self.xml.get_widget("PackageDescription")
        DetailsButton = self.xml.get_widget("PackageSelectionDetailsButton")
        ToggleLabel = self.xml.get_widget("PackageSelectLabel")
        MediaSelectToggle = self.xml.get_widget("MediaSelectToggle")

        if self.installTask.profile:
            self.packageTreeModel.foreach(self.select_package_by_appid, self.installTask.profile.appid)

            if self.installTask.profile.app_profile.is_supported:
                supportedLabel.set_markup("<b>"+_("Supported by CodeWeavers")+"</b>")
            else:
                supportedLabel.set_markup("<b>"+_("Not Supported by CodeWeavers")+"</b>")

            medal = self.installTask.profile.app_profile.medal
            if medal:
                medalImage.set_from_file(self.installTask.profile.app_profile.medal_icon)
                markup = '<b>%s</b>\n%s' % (self.medal_names[medal], self.medal_descriptions[medal])
            else:
                medalImage.set_from_file(None)
                markup = ''

            if hasattr(medalImage, 'set_tooltip_markup'):
                # GTK >= 2.12
                medalImage.set_tooltip_markup(markup)

            description = "<body>\n%s</body>" % (self.installTask.profile.app_profile.description)
            descriptionTextView.get_buffer().set_text("")
            try:
                #  Encode description as utf8 because htmltextview likes that better.
                descriptionTextView.display_html(description.encode('utf8'))
            except:
                cxlog.log("Display_html is failing to render the string %s.  Probably a malformed string is being displayed." % cxlog.debug_str(description))
                descriptionTextView.get_buffer().set_text(self.installTask.profile.app_profile.description)

            if self.installTask.profile.details_url:
                DetailsButton.set_sensitive(True)
            else:
                DetailsButton.set_sensitive(False)

            ToggleLabel.set_text(_("Will install %s") % self.installTask.profile.name)

        else:
            packageTreeView = self.xml.get_widget("PackageTreeView")
            packageTreeView.get_selection().unselect_all()
            supportedLabel.set_markup("")
            medalImage.set_from_file(None)
            descriptionTextView.get_buffer().set_text("")
            DetailsButton.set_sensitive(False)
            ToggleLabel.set_text(_("Select an application to install"))

        if self.installTask.virtual:
            MediaSelectToggle.hide()
        else:
            MediaSelectToggle.show()

        self.xml.get_widget("BottleSelectProgress").show()
        self.xml.get_widget("BottleSelectListing").hide()

        self.updateMediaListing()
        self.update_common_widgets()


    def details_clicked(self, _button):
        packageTreeView = self.xml.get_widget("PackageTreeView")
        cursor = packageTreeView.get_cursor()
        selectedID = self.packageTreeModel.get_value(self.packageTreeModel.get_iter(cursor[0]), 1)
        selectedPackage = self.installTask.profiles[selectedID]
        cxutils.launch_url(selectedPackage.details_url)

    def count_packages(self, model, root_path, treeiter, result=None):
        if result is None:
            result = {}

        count = 0

        while treeiter is not None:
            is_category = model.get_value(treeiter, 3)

            if is_category:
                _result, subcount = self.count_packages(model, model.get_path(treeiter), model.iter_children(treeiter), result)
                count += subcount
            else:
                count += 1

            treeiter = model.iter_next(treeiter)

        result[root_path] = count

        return result, count

    def package_select_search_changed(self, _caller):
        searchWidget = self.xml.get_widget("PackageSelectSearch")
        self.packageSearchString = searchWidget.get_text()
        self.packageTreeModel.refilter()

        package_counts, _root_count = self.count_packages(self.packageTreeModel, (), self.packageTreeModel.get_iter_first())

        # Check whether there's 1 category that's not the root or unsupported
        # that has items, and expand it if so.
        categories_with_items = []
        for root_path in package_counts:
            if root_path != ():
                treeiter = self.packageTreeModel.get_iter(root_path)
                category_id = self.packageTreeModel.get_value(treeiter, 1)
                if category_id != '.unsupported' and package_counts[root_path]:
                    categories_with_items.append(root_path)

        if len(categories_with_items) == 1:
            self.xml.get_widget('PackageTreeView').expand_row(categories_with_items[0], False)

    #-----------------------------------------------------------------------
    #  Installer Media Selection
    #-----------------------------------------------------------------------

    def browse_for_installer_file(self, caller):
        filePicker = gtk.FileChooserDialog(_("Choose a Windows installer file:"), caller, gtk.FILE_CHOOSER_ACTION_OPEN)

        cxguitools.add_filters(filePicker, cxguitools.FILTERS_INSTALLABLE)
        filePicker.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        filePicker.add_button(gtk.STOCK_OPEN, gtk.RESPONSE_OK)

        if filePicker.run() == gtk.RESPONSE_OK:
            selectedFile = filePicker.get_filename()
        else:
            selectedFile = None

        filePicker.destroy()

        return selectedFile


    def browse_for_installer_dir(self, caller):
        filePicker = gtk.FileChooserDialog(_("Choose the directory containing your installer."), caller, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)

        filePicker.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        filePicker.add_button(gtk.STOCK_OPEN, gtk.RESPONSE_OK)

        if filePicker.run() == gtk.RESPONSE_OK:
            selectedDir = filePicker.get_filename()
        else:
            selectedDir = None

        filePicker.destroy()

        return selectedDir


    def volumeAdded(self, volume):
        if volume.mountpoint == "/" or not volume.is_disc:
            # unlikely that the user wants to use this.
            return

        # We need to check this is not a duplicate because volumeAdded() is
        # also used for the user-selected installers and directories
        volumesIter = self.volumesListModel.iter_children(None)
        while volumesIter != None:
            if self.volumesListModel.get_value(volumesIter, 2) == volume.mountpoint:
                # This is a duplicate so we have nothing to do
                return
            volumesIter = self.volumesListModel.iter_next(volumesIter)

        if volume.label:
            label = volume.label
        else:
            label = os.path.basename(volume.mountpoint)
        volumesWidget = self.xml.get_widget("MediaTreeView")
        self.installTask.source_medias[volume.mountpoint] = installtask.SourceMedia(self.installTask, volume.mountpoint, volume.label)

        newrow = self.volumesListModel.append()
        self.volumesListModel.set_value(newrow, 0, label)
        self.volumesListModel.set_value(newrow, 2, volume.mountpoint)
        if volume.is_disc:
            pixbuf = volumesWidget.render_icon(gtk.STOCK_CDROM, gtk.ICON_SIZE_DIALOG)
        else:
            pixbuf = volumesWidget.render_icon(gtk.STOCK_HARDDISK, gtk.ICON_SIZE_DIALOG)
        self.volumesListModel.set_value(newrow, 1, pixbuf)
        self.volumesListModel.set_value(newrow, 3, '0'+self.volumesListModel.get_value(newrow, 0))


    def volumeDeleted(self, volume):
        del self.installTask.source_medias[volume.mountpoint]

        volumesIter = self.volumesListModel.iter_children(None)
        while volumesIter != None:
            if self.volumesListModel.get_value(volumesIter, 2) == volume.mountpoint:
                self.volumesListModel.remove(volumesIter)
                break
            else:
                volumesIter = self.volumesListModel.iter_next(volumesIter)
        else:
            # currentVolumeLabel not found
            cxlog.log("Warning: removed volume does not exist: %s" % cxlog.debug_str(volume))


    def wantMediaSelection(self):
        return not (self.installTask.GetInstallerSource() or
                    self.installTask.installerDownloadSource or
                    self.installTask.virtual)

    def media_toggle(self, _widget):
        if self.xml.get_widget('MediaSelectContent').get_property('visible'):
            self.advance_section()
        else:
            self.set_expanded_section('MediaSelect')


    chooseInstallerFileString = _(u"Choose Installer File\u2026")
    chooseInstallerFolderString = _(u"Choose Installer Folder\u2026")
    def updateMediaListing(self):
        # Do we want to display the download options?
        downloadRadio = self.xml.get_widget("DownloadRadio")
        localRadio = self.xml.get_widget("LocalRadio")
        downloadFrame = self.xml.get_widget("DownloadFrame")
        app_profile = self.installTask.profile and self.installTask.profile.app_profile

        if app_profile and (app_profile.download_page_urls or app_profile.download_urls):
            downloadCombo = self.xml.get_widget("DownloadCombo")

            downloadRadio.show()
            localRadio.show()
            downloadFrame.show()
            self.urlListStore.clear()

            if app_profile.download_urls:
                downloadCombo.show()
                # Sort the localized languages names in alphabetical order.
                name2lang = {'': ''}
                for lang in app_profile.download_urls:
                    if lang == '':
                        # We map the default to an empty string so it ends up
                        # as the first entry in the sorted list. This also
                        # saves us trouble with localization.
                        continue
                    lang_name = c4profiles.LANGUAGES.get(lang, lang)
                    if lang_name in name2lang:
                        # This is not expected to happen and if it did it
                        # would still mean the user would have no way of
                        # making a difference between the two locales. So it
                        # is not supported at all.
                        cxlog.err("%s and %s both map to %s" % (cxlog.debug_str(name2lang[lang_name]), cxlog.debug_str(lang), cxlog.debug_str(lang_name)))
                        continue
                    name2lang[lang_name] = lang
                for lang_name in sorted(name2lang):
                    lang = name2lang[lang_name]
                    if lang_name == '':
                        lang_name = c4profiles.LANGUAGES.get('', '')
                    newrow = self.urlListStore.append()
                    self.urlListStore.set_value(newrow, 0, "%s: %s" % (lang_name, app_profile.download_urls[lang]))
                    self.urlListStore.set_value(newrow, 1, app_profile.download_urls[lang])
                    self.urlListStore.set_value(newrow, 2, lang)
                # If we get notification that the source changed before the package, we have to update the selection now
                self.urlListStore.foreach(self.select_download, None)
            else:
                downloadCombo.hide()

            
            downloadNotes = self.xml.get_widget("DownloadInstallationNotes")
            if app_profile.download_page_urls:
                _lang, url = cxutils.get_language_value(app_profile.download_page_urls)
                if app_profile.download_urls:
                    # FIXME: Show the download URL that matches the currently
                    # selected download url.
                    # FIXME: If there's no download url, then the user won't
                    # be able to change the download page url.
                    note = _('Additional info and/or installers can be found at <a href="%(url)s">%(label)s</a>.') % {'url': cxutils.html_escape(url), 'label': cxutils.html_escape(url)}
                else:
                    note = _('An installer can be found at <a href="%(url)s">%(label)s</a>.') % {'url': cxutils.html_escape(url), 'label': cxutils.html_escape(url)}

                fullNote = "<body> %s </body>" % note
                downloadNotes.get_buffer().set_text("")
                try:
                    downloadNotes.display_html(fullNote.encode('utf8'))
                except:
                    cxlog.warn("Display_html is failing to render the string %s.\n%s." % (cxlog.debug_str(fullNote), traceback.format_exc()))
                    downloadNotes.get_buffer().set_text(fullNote)
                downloadNotes.show()
            else:
                downloadNotes.hide()
            
        else:
            downloadRadio.hide()
            localRadio.hide()
            downloadFrame.hide()

    
    def select_media_by_mountpoint(self, model, path, treeiter, mountpoint):
        if model.get_value(treeiter, 2) == mountpoint:
            self.xml.get_widget("MediaTreeView").set_cursor(path)
            self.xml.get_widget("MediaTreeView").select_path(path)
            return True # stop iteration
        else:
            return False # continue iteration

    def select_download(self, model, _path, treeiter, _userdata):
        if model.get_value(treeiter, 1) == self.installTask.installerDownloadSource and \
           model.get_value(treeiter, 2) == self.installTask.installerDownloadLocale:
            self.xml.get_widget("DownloadCombo").set_active_iter(treeiter)
            return True # stop iteration
        else:
            return False # continue iteration

    def sourceChanged(self):
        downloadRadio = self.xml.get_widget("DownloadRadio")
        downloadFrame = self.xml.get_widget("DownloadFrame")
        localRadio = self.xml.get_widget("LocalRadio")
        localFrame = self.xml.get_widget("VolumeFrame")
        ToggleLabel = self.xml.get_widget("MediaSelectLabel")

        if self.installTask.installerSource:
            self.volumesListModel.foreach(self.select_media_by_mountpoint, self.installTask.installerSource)

            localRadio.set_active(True)
            localFrame.set_sensitive(True)
            downloadRadio.set_active(False)
            downloadFrame.set_sensitive(False)

            ToggleLabel.set_label(_("Will install from %s") % self.installTask.installerSource)

        elif self.installTask.installerDownloadSource:
            self.urlListStore.foreach(self.select_download, None)

            localRadio.set_active(False)
            localFrame.set_sensitive(False)
            downloadRadio.set_active(True)
            downloadFrame.set_sensitive(True)

            if self.installTask.installerDownloadLocale:
                lang = self.installTask.installerDownloadLocale
                lang_name = c4profiles.LANGUAGES.get(lang, lang)
                ToggleLabel.set_label(_("Will download the %(language)s installer from %(url)s") % {'language': lang_name, 'url': self.installTask.installerDownloadSource})
            else:
                ToggleLabel.set_label(_("Will download the installer from %s") % self.installTask.installerDownloadSource)

        else:
            self.xml.get_widget("MediaTreeView").unselect_all()

            localRadio.set_active(True)
            localFrame.set_sensitive(True)
            downloadRadio.set_active(False)
            downloadFrame.set_sensitive(False)

            ToggleLabel.set_label(_("Select an installer"))

        self.update_common_widgets()


    def media_selection_changed(self):
        downloadRadio = self.xml.get_widget("DownloadRadio")
        localRadio = self.xml.get_widget("LocalRadio")

        if downloadRadio.get_active():
            downloadCombo = self.xml.get_widget("DownloadCombo")
            activeIndex = downloadCombo.get_active_iter()
            if activeIndex is None:
                # this can happen while we're updating the combo list model
                return
            selectedURL = self.urlListStore.get_value(activeIndex, 1)
            selectedLocale = self.urlListStore.get_value(activeIndex, 2)
            cxlog.log("selectedURL: %s" % cxlog.debug_str(selectedURL))
            self.installTask.SetInstallerDownload(selectedURL, selectedLocale)

        elif localRadio.get_active():
            mediaTreeView = self.xml.get_widget("MediaTreeView")
            selections = mediaTreeView.get_selected_items()

            if selections:
                selectedPath = selections[0]
                selectedIter = self.volumesListModel.get_iter(selectedPath)
                volumeName = self.volumesListModel.get_value(selectedIter, 0)
                mountPoint = self.volumesListModel.get_value(selectedIter, 2)

                if volumeName == self.chooseInstallerFileString:
                    selectedFile = self.browse_for_installer_file(self.xml.get_widget("InstallerAssistant"))
                    if selectedFile:
                        volume = mountpoints.Volume()
                        volume.mountpoint = selectedFile
                        volume.is_disc = True # Lie so it's taken into account
                        self.volumeAdded(volume)
                        self.installTask.SetInstallerSource(selectedFile)
                    else:
                        self.installTask.ClearInstallerSource()
                        mediaTreeView.unselect_all()
                elif volumeName == self.chooseInstallerFolderString:
                    selectedDir = self.browse_for_installer_dir(self.xml.get_widget("InstallerAssistant"))
                    if selectedDir:
                        volume = mountpoints.Volume()
                        volume.mountpoint = selectedDir
                        volume.is_disc = True # Lie so it's taken into account
                        self.volumeAdded(volume)
                        self.installTask.SetInstallerSource(selectedDir)
                    else:
                        self.installTask.ClearInstallerSource()
                        mediaTreeView.unselect_all()
                else:
                    self.installTask.SetInstallerSource(mountPoint)
            else:
                self.installTask.ClearInstallerSource()

    def download_changed_gui(self, _widget):
        if self.xml.get_widget("DownloadRadio").get_active():
            self.media_selection_changed()

    def local_changed_gui(self, _widget):
        if self.xml.get_widget("LocalRadio").get_active():
            self.media_selection_changed()

    #-----------------------------------------------------------------------
    #  Bottle Selection
    #-----------------------------------------------------------------------

    category_markup = {
        installtask.CAT_NONE: "<b>"+_("Bottles")+"</b>",
        installtask.CAT_RECOMMENDED: "<b>"+_("Recommended Bottles")+"</b>",
        installtask.CAT_COMPATIBLE: "<b>"+_("Compatible Bottles")+"</b>",
        installtask.CAT_INCOMPATIBLE: "<b>"+_("Incompatible Bottles")+"</b>"
        }

    category_sort = {
        installtask.CAT_NONE: '0',
        installtask.CAT_RECOMMENDED: '1',
        installtask.CAT_COMPATIBLE: '2',
        installtask.CAT_INCOMPATIBLE: '3',
        }

    def categorizedBottle_(self, target_bottle):
        pass

    def categorizedAllBottles(self):
        self.updateBottleListing()

    def analyzedBottle_(self, target_bottle):
        if target_bottle is self.installTask.target_bottle:
            self.update_common_widgets()

    def invalidNewBottleName(self):
        cxlog.log("invalid new bottle name [%s]" % cxlog.to_str(self.installTask.newBottleName))
        # FIXME: Todo

    def wantBottleSelection(self):
        if not self.installTask.bottlename:
            # no bottle selected
            return True
        elif self.installTask.GetCreateNewBottle():
            # check that the name doesn't match an existing bottle
            for bottlename in bottlecollection.sharedCollection().bottleList():
                if self.installTask.bottlename.lower() == bottlename.lower():
                    return True
            else:
                return False


    def bottle_toggle(self, _widget):
        if self.xml.get_widget('BottleSelectContent').get_property('visible'):
            self.advance_section()
        else:
            self.set_expanded_section('BottleSelect')


    def update_bottle_analysis_progress(self):
        bottles = bottlecollection.sharedCollection().bottles()
        total = len(bottles)
        bottles_completed = 0

        if total:
            for bottleobj in bottles:
                if bottleobj.installed_packages_ready:
                    bottles_completed += 1

            progressbar = self.xml.get_widget("BottleSelectProgress")
            progressbar.set_fraction(float(bottles_completed) / total)

        if bottles_completed == total:
            # stop the timer
            self.bottle_progress_source = None
            return False
        else:
            # continue
            return True

    def update_bottle_selection_label(self):
        ToggleLabel = self.xml.get_widget("BottleSelectLabel")
        bottlename = self.installTask.bottlename
        if not bottlename:
            ToggleLabel.set_label(_("Select a bottle into which to install"))
        elif self.installTask.GetCreateNewBottle():
            ToggleLabel.set_label(_("Will install into a new %(template)s bottle '%(name)s'") % {
                'template': self.installTask.newBottleTemplate,
                'name': bottlename})
        else:
            ToggleLabel.set_label(_("Will install into the bottle '%(name)s'") % {'name': bottlename})

    def select_bottle(self, model, path, treeiter, _userdata):
        if self.installTask.GetCreateNewBottle():
            bottle_string = '.' + self.installTask.GetNewBottleTemplate()
        elif self.installTask.bottlename:
            bottle_string = self.installTask.bottlename
        else:
            bottle_string = None

        if model.get_value(treeiter, 0) == bottle_string:
            self.xml.get_widget("BottleSelectionView").set_cursor(path)
            return True # stop iteration
        else:
            return False # continue iteration

    def updateBottleListing(self):
        self.bottleListModel.clear()

        bottles_by_category = {}

        for target_bottle in self.installTask.bottles.values() + self.installTask.templates.values():
            category = target_bottle.get_category()
            if category in bottles_by_category:
                bottles_by_category[category].append(target_bottle)
            else:
                bottles_by_category[category] = [target_bottle]

        for category in bottles_by_category:
            category_row = self.bottleListModel.append(None, row=('', self.category_markup[category], '', self.category_sort[category]))
            for target_bottle in bottles_by_category[category]:
                if isinstance(target_bottle, installtask.TargetTemplate):
                    name = '.' + target_bottle.template
                    markup = _("(New %(template)s bottle)") % {'template': target_bottle.template}
                    sortkey = '0' + target_bottle.template
                else:
                    name = target_bottle.bottlename
                    markup = target_bottle.bottlename
                    sortkey = '1' + target_bottle.bottlename
                template = target_bottle.template
                self.bottleListModel.append(parent=category_row, row=(name, markup, template, sortkey))

        self.xml.get_widget("BottleSelectionView").expand_all()
        self.bottleListModel.foreach(self.select_bottle, None)

        self.xml.get_widget("BottleSelectProgress").hide()
        self.xml.get_widget("BottleSelectListing").show()


    def bottleCreateChanged(self):
        self.bottleListModel.foreach(self.select_bottle, None)
        if self.installTask.GetCreateNewBottle():
            self.xml.get_widget("NewBottleWidgets").show()
        else:
            self.xml.get_widget("NewBottleWidgets").hide()
        self.update_bottle_selection_label()
        self.update_common_widgets()

    def bottleNameChanged(self):
        self.bottleListModel.foreach(self.select_bottle, None)
        self.update_bottle_selection_label()
        self.update_common_widgets()

    def bottleNewnameChanged(self):
        if self.installTask.newBottleName:
            self.xml.get_widget("NewBottleNameEntry").set_text(self.installTask.newBottleName)
        self.update_bottle_selection_label()
        self.update_common_widgets()

    def bottleTemplateChanged(self):
        self.bottleListModel.foreach(self.select_bottle, None)
        self.update_bottle_selection_label()
        self.update_common_widgets()

    def bottle_selection_changed_gui(self, _widget):
        bottlesWidget = self.xml.get_widget("BottleSelectionView")
        tmodel, titer = bottlesWidget.get_selection().get_selected()
        if titer is None:
            # This can happen while we're rebuilding the bottle list
            return
        selectedBottleName = tmodel.get_value(titer, 0)

        if selectedBottleName == '':
            # For categories, select the first child.
            self.xml.get_widget("BottleSelectionView").set_cursor(tmodel.get_path(tmodel.iter_children(titer)))
        elif selectedBottleName.startswith('.'):
            # A bottle template.
            self.installTask.SetCreateNewBottle(create=True, template=selectedBottleName[1:])
        else:
            self.installTask.bottlename = selectedBottleName

    def bottle_newname_changed_gui(self, _widget):
        self.installTask.SetCreateNewBottle(
            newbottlename=self.xml.get_widget("NewBottleNameEntry").get_text())

    def new_bottle_insert_text(self, caller, new_text, position, user_data):
        filtered = cxutils.bottlename_sanitize(new_text)
        if filtered != new_text:
            caller.emit_stop_by_name("insert-text")
            caller.emit("insert-text", filtered, position, user_data)


    def no_section_activated(self):
        if self.installEngine:
            # Install was started already.
            return

        self.update_common_widgets()
    

    #-----------------------------------------------------------------------
    #  Install
    #-----------------------------------------------------------------------


    def doInstall(self):
        self.tasks = {}
        self.cancelInstallation = False
        self.install_in_progress = True
        self.installEngine = cxaiengine.Engine(self.installTask)

        self.totalTasks = len(self.installEngine.get_sorted_tasks()) 
        self.completedTasks = 0
        self.someTasksCancelled = False
        self.someTasksSkipped = False

        self.spoolAvailableTasks()


    def displayInstallationProgressPanel(self):
        self.xml.get_widget("PackageSelectToggle").set_sensitive(False)
        self.xml.get_widget("MediaSelectToggle").set_sensitive(False)
        self.xml.get_widget("BottleSelectToggle").set_sensitive(False)
        self.xml.get_widget("InstallerNext").set_sensitive(False)
        self.xml.get_widget("InstallerToggle").show()
        label = self.xml.get_widget("InstallerLabel")
        label.set_text(_(u"Installing %s\u2026") % self.installTask.profile.name)
        self.set_expanded_section('Installer')
        self.update_install_notes(False)

    def continueInstallation(self):
        if not self.install_in_progress:
            return

        progbar = self.xml.get_widget("installerProgbar")
        self.completedTasks += 1.0
        fraction = self.completedTasks/self.totalTasks
        progbar.set_fraction(fraction)

        if self.cancelInstallation:
            cxlog.log("Cancelled Installation")

            self.install_in_progress = False

            self.displayInstallationFinishedPanel(True)
        elif self.installEngine.all_done():
            cxlog.log("Installation finished")

            self.install_in_progress = False

            self.displayInstallationFinishedPanel(False)
        else:
            self.spoolAvailableTasks()

    def scroll_tasks(self):
        tasksScroll = self.xml.get_widget("InstallerTasksScroll")
        adj = tasksScroll.get_vadjustment()
        adj.set_value(adj.upper)

    def spoolAvailableTasks(self):
        # Show a failure dialog if necessary
        if self.failing_tasks and not self.install_ui:
            self.failing_tasks.pop(0).show_failure_dialog()
        # Add all the tasks that are currently runnable to the operation queue.
        for task in sorted(self.installEngine.runnable(), cmp=cxaiebase.tasks_cmp):
            if not self.install_ui or not isinstance(task, interactive_tasks):
                if self.installEngine.schedule(task):
                    self.tasks[task] = InstallTask(task, self)

    def add_failing_task(self, task):
        self.failing_tasks.append(task)
        self.continueInstallation()

    def cancel_install(self):
        self.cancelInstallation = True
        self.someTasksCancelled = True
        for task in self.tasks.itervalues():
            task.cancel()
        self.continueInstallation()

    def install_toggle(self, _widget):
        if self.xml.get_widget('InstallerContent').get_property('visible'):
            self.set_expanded_section(None)
        else:
            self.set_expanded_section('Installer')

    def install_section_activated(self):
        pass


    def displayInstallationFinishedPanel(self, inCancelled):
        if inCancelled:
            titleString = _("Installation Canceled")
        else:
            titleString = _("Installation Complete")

        completionMessage = ""
        if self.someTasksCancelled:
            completionMessage = _("The installation failed.")
        elif self.someTasksSkipped:
            completionMessage = _("The installation completed, but some tasks were skipped.")
        else:
            completionMessage = _("The installation completed.")

        if 'logfile' in self.installEngine.state:
            url = self.installEngine.state["logfile"]
        elif 'CX_LOG' in self.installEngine.state["environ"]:
            url = self.installEngine.state["environ"]["CX_LOG"]
        else:
            url = None

        if url:
            # FIXME: We should escape url in case it contains weird characters
            # such as '<', '>', '"', etc. Otherwise we will get HTML errors.
            captionHTML = _("<body> <p> %(completion)s </p> <p> More diagnostic information can be found in the <a href=\"file://%(url)s\">debug log</a>. </p> </body>") % {'completion': completionMessage, 'url': url}
        else:
            captionHTML = "<body> <p> %s </p> </body>" % completionMessage

        try:
            #  Encode description as utf8 because htmltextview likes that better.
            caption = self.xml.get_widget("InstallResults")
            caption.display_html(captionHTML.encode('utf8'))
        except:
            cxlog.log("Display_html is failing to render the string %s.  Probably a malformed string is being displayed." % cxlog.debug_str(captionHTML))
            caption.get_buffer().set_text(captionHTML)

        self.xml.get_widget("FinishedLabel").set_text(titleString)
        self.xml.get_widget("FinishedToggle").show()
        self.xml.get_widget("InstallerCancel").hide()
        self.set_expanded_section("Finished")

        nextButton = self.xml.get_widget("InstallerNext")
        nextButton.set_sensitive(True)
        stock_close = gtk.stock_lookup(gtk.STOCK_CLOSE)
        if stock_close:
            _stock_id, label, _modifier, _keyval, _translation_domain = stock_close
            nextButton.set_label(label)

        try:
            # Touch the config file so the bottle manager knows that
            #  something is happening.
            if self.installTask.createNewBottle:
                os.utime(bottlequery.config_file_path(self.installTask.newBottleName), None)
            else:
                os.utime(bottlequery.config_file_path(self.installTask.existingBottleName), None)
        except bottlequery.NotFoundError:
            # The bottle wasn't created yet. Do nothing.
            pass

    def finished_toggle(self, _widget):
        if self.xml.get_widget('FinishedContent').get_property('visible'):
            self.set_expanded_section(None)
        else:
            self.set_expanded_section('Finished')

    def finished_section_activated(self):
        pass


    #Used to update global state (such as the "Proceed" button name/sensitivity)
    # when the active section changes.
    on_section_changed = {
        'PackageSelect': update_common_widgets,
        'MediaSelect': update_common_widgets,
        'BottleSelect': update_common_widgets,
        'Installer': install_section_activated,
        'Finished': finished_section_activated,
        None: no_section_activated,
    }

    def show_help(self, _caller):
        cxguitools.show_help("install-windows.html")

    def animate(self):
        for entry in self.animateFunctions:
            function, data = entry
            if data:
                function(data)
            else:
                function()
        return True

    def addAnimationFunction(self, inFunction, userData = None):
        self.animateFunctions.append([inFunction, userData])

    def removeAnimationFunction(self, inFunction, userData = None):
        if inFunction in self.animateFunctions:
            self.animateFunctions.remove([inFunction, userData])

def update_progress_bar(inData):
    task, progbar = inData
    if isinstance(task, cxaiemedia.AIEDownload):
        if task.size == -1:
            progbar.set_text(_(u"Preparing to download\u2026"))
            progbar.pulse()
        elif task.size == 0:
            progbar.set_text(_("(unknown size)"))
            progbar.pulse()
        else:
            def size_to_str(size):
                if size < 2 * 1024 * 1024:
                    return _("%dkB") % (size / 1024)
                else:
                    return _("%.1fMB") % (size / 1024.0 / 1024)

            progbar.set_text(_("Downloaded %(part)s of %(total)s") % {'part':size_to_str(task.downloaded), 'total':size_to_str(task.size)})
            progbar.set_fraction(task.progress)

interactive_tasks = (cxaiecore.AIECore, )

class InstallTask(object):
    # widgets
    icon = None
    caption = None
    progbar = None
    frame = None
    cancel_button = None
    finished = False

    def __init__(self, aietask, wizard):
        self.aietask = aietask
        self.wizard = wizard

        if aietask.label:
            tasksBox = self.wizard.xml.get_widget("InstallerTasksBox")
            taskFrame = gtk.VBox()
            taskBox = gtk.HBox()

            progbar = gtk.ProgressBar()
            if isinstance(aietask, cxaiemedia.AIEDownload):
                icon = gtk.image_new_from_stock(gtk.STOCK_COPY, gtk.ICON_SIZE_DND)
                progbar.set_pulse_step(0.001)
                self.wizard.addAnimationFunction(update_progress_bar, [aietask, progbar])
            elif isinstance(aietask, cxaiemedia.AIEScanMedia):
                icon = gtk.image_new_from_stock(gtk.STOCK_FIND, gtk.ICON_SIZE_DND)
                progbar.set_pulse_step(0.02)
                self.wizard.addAnimationFunction(progbar.pulse)
            else:
                icon = gtk.image_new_from_stock(gtk.STOCK_EXECUTE, gtk.ICON_SIZE_DND)
                progbar.set_pulse_step(0.03)
                self.wizard.addAnimationFunction(progbar.pulse)
            taskBox.pack_start(icon, False, False, 2)
            self.icon = icon

            progBox = gtk.VBox()
            label1 = gtk.Label(_(u"%s\u2026") % aietask.label)
            label1.set_alignment(0, 0.5)
            self.caption = label1
            progBox.pack_start(label1, False, False, 2)
            self.progbar = progbar

            progBox.pack_start(progbar, False, True, 2)
            taskBox.pack_start(progBox, True, True, 10)

            self.cancel_button = gtk.Button(stock=gtk.STOCK_CANCEL)
            self.cancel_button.connect('clicked', self.cancel_clicked)
            cancel_vbox = gtk.VBox()
            cancel_vbox.pack_start(gtk.Label())
            cancel_vbox.pack_start(self.cancel_button, False, False)
            cancel_vbox.pack_start(gtk.Label())
            taskBox.pack_start(cancel_vbox, False, False)

            self.frame = taskFrame
            taskFrame.pack_start(taskBox, False, False, 2)
            taskFrame.pack_start(gtk.HSeparator(), False, False, 2)
            tasksBox.pack_start(taskFrame, False, False, 0)
            taskFrame.show_all()
            tasksBox.show()

            self.wizard.scroll_tasks()

        self.enqueue()

    def enqueue(self):
        if isinstance(self.aietask, interactive_tasks):
            self.wizard.install_ui = True
        self.operation = InstallTaskOperation(self)
        pyop.sharedOperationQueue.enqueue(self.operation)

    def done(self):
        # This task has either succeeded or failed and will not retry. Update
        # widgets and notify the ai engine.
        if self.progbar:
            self.progbar.hide()
        if self.cancel_button:
            self.cancel_button.hide()
        if isinstance(self.aietask, interactive_tasks):
            self.wizard.install_ui = False

        self.finished = True
        self.wizard.scroll_tasks()
        self.aietask.done()
        self.wizard.continueInstallation()

    def task_returned(self, success):
        if self.finished:
            return
        if isinstance(self.aietask, interactive_tasks):
            self.wizard.install_ui = False
        self.operation = None
        if success:
            if self.caption:
                self.caption.set_text(self.caption.get_text() + " " + _("Done."))
            self.done()
        else:
            if self.caption:
                self.caption.set_text(self.caption.get_text() + " " + _("Failed."))

            self.wizard.add_failing_task(self)

    def show_failure_dialog(self):
        label = self.aietask.label
        if not label:
            label = unicode(self.aietask)

        if self.aietask.error:
            failString = _("An error occurred while running the '%(task)s' task:\n\n%(error)s") % {'task': label, 'error': self.aietask.error}
        else:
            failString = _("An unexpected error occurred while running the '%(task)s' task.") % {'task':label}

        panel = self.wizard.xml.get_widget("InstallationProgress")
        cxguitools.CXMessageDlg(failString,
                                title="%s error" % label,
                                button_array=[[_("Try Again"), 1],
                                              [_("Skip This Step"), 0],
                                              [_("Cancel Installation"), 2]],
                                response_function=self.abortRetryFail,
                                parent=panel)

        self.wizard.install_ui = True

    def skip(self):
        self.wizard.someTasksSkipped = True
        self.done()

    def retry(self):
        if self.caption:
            self.caption.set_text(self.caption.get_text() + u' ' + _(u"Retrying\u2026"))
        self.enqueue()

    def cancel(self):
        if self.finished:
            return
        if self.aietask.can_cancel():
            self.aietask.cancel()
        if self.caption:
            self.caption.set_text(self.caption.get_text() + u' ' + _("Canceled."))
        self.done()

    def abortRetryFail(self, response):
        self.wizard.install_ui = False
        if response == 0:
            self.skip()
        elif response == 1:
            self.retry()
        elif response == 2:
            self.wizard.cancel_install()
        self.wizard.continueInstallation() # Just in case we need to show another failure dialog

    def cancel_clicked(self, _widget):
        self.cancel()

class InstallTaskOperation(pyop.PythonOperation):

    def __init__(self, task):
        pyop.PythonOperation.__init__(self)
        self.task = task
        self.success = False

    def __unicode__(self):
        return "%s - %s" % (self.__class__.__name__, unicode(self.task))

    def main(self):
        if not self.task.wizard.cancelInstallation and not self.task.finished:
            self.success = self.task.aietask.main()
        else:
            self.success = False

    def finish(self):
        self.task.task_returned(self.success)

